import textwrap
from typing import List as TList

import discord
from cpl_discord.command import DiscordCommandABC
from cpl_query.extension import List
from cpl_translation import TranslatePipe
from discord import app_commands
from discord.ext import commands
from discord.ext.commands import Context

from bot_core.abc.client_utils_service_abc import ClientUtilsServiceABC
from bot_core.abc.message_service_abc import MessageServiceABC
from bot_core.logging.command_logger import CommandLogger
from bot_data.abc.server_repository_abc import ServerRepositoryABC
from modules.permission.abc.permission_service_abc import PermissionServiceABC
from modules.stats.model.statistic import Statistic
from modules.stats.service.statistic_service import StatisticService
from modules.stats.ui.add_statistic_form import AddStatisticForm

stats = List(Statistic, [
    Statistic(
        'Benutzer XP Aufsteigend',
        'Zeigt XP von jedem Benutzer, aufsteigend nach XP',
        textwrap.dedent("""\
        result.header.append('Name')
        result.header.append('XP')
        
        for user in users.order_by(lambda u: u.xp):
            row = List(str)
            member = guild.get_member(user.discord_id)
            row.append(member.name)
            row.append(str(user.xp))
            result.values.append(row)
        """)),
    Statistic('Benutzer XP Absteigend', 'Zeigt XP von jedem Benutzer, absteigend nach XP', '')
])


class StatsGroup(DiscordCommandABC):

    def __init__(
            self,
            logger: CommandLogger,
            message_service: MessageServiceABC,
            client_utils: ClientUtilsServiceABC,
            translate: TranslatePipe,
            permission_service: PermissionServiceABC,
            statistic: StatisticService,
            servers: ServerRepositoryABC,
    ):
        DiscordCommandABC.__init__(self)

        self._logger = logger
        self._client_utils = client_utils
        self._message_service = message_service
        self._t = translate
        self._permissions = permission_service
        self._statistic = statistic
        self._servers = servers

    @commands.hybrid_group()
    @commands.guild_only()
    async def stats(self, ctx: Context):
        pass

    @stats.command()
    @commands.guild_only()
    async def list(self, ctx: Context, wait: int = None):
        self._logger.debug(__name__, f'Received command stats list {ctx}')
        if not await self._client_utils.check_if_bot_is_ready_yet_and_respond(ctx):
            return

        if not self._permissions.is_member_moderator(ctx.author):
            await self._message_service.send_ctx_msg(ctx, self._t.transform('common.no_permission_message'))
            self._logger.trace(__name__, f'Finished command stats list')
            return

        embed = discord.Embed(
            title=self._t.transform('modules.auto_role.list.title'),
            description=self._t.transform('modules.auto_role.list.description'),
            color=int('ef9d0d', 16)
        )

        statistics = ''
        descriptions = ''
        for statistic in stats:
            statistics += f'\n{statistic["Name"]}'
            descriptions += f'\n{statistic["Description"]}'

        embed.add_field(name=self._t.transform('modules.stats.list.statistic'), value=statistics, inline=True)
        embed.add_field(name=self._t.transform('modules.stats.list.description'), value=statistics, inline=True)
        await self._message_service.send_ctx_msg(ctx, embed, wait_before_delete=wait)
        self._logger.trace(__name__, f'Finished command stats list')

    @stats.command()
    @commands.guild_only()
    async def view(self, ctx: Context, name: str, wait: int = None):
        self._logger.debug(__name__, f'Received command stats {ctx}:{name}')
        if not await self._client_utils.check_if_bot_is_ready_yet_and_respond(ctx):
            return

        if ctx.guild is None:
            return

        try:
            server = self._servers.get_server_by_discord_id(ctx.guild.id)
            statistic = stats.where(lambda s: s.name == name).single()
            result = await self._statistic.execute(statistic.code, server)

            embed = discord.Embed(
                title=statistic.name,
                description=statistic.description,
                color=int('ef9d0d', 16)
            )

            for i in range(result.header.count()):
                header = result.header[i]
                value = ''
                for row in result.values:
                    value += f'\n{row[i]}'
                embed.add_field(name=header, value=value, inline=True)

            await self._message_service.send_ctx_msg(ctx, embed, wait_before_delete=wait)
        except Exception as e:
            self._logger.error(__name__, f'Cannot view statistic {name}', e)
            await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.stats.view.failed'))

        self._logger.trace(__name__, f'Finished stats command')

    @view.autocomplete('name')
    async def view_autocomplete(self, interaction: discord.Interaction, current: str) -> TList[app_commands.Choice[str]]:
        return [app_commands.Choice(name=f'{statistic.name}: {statistic.description}', value=statistic.name) for statistic in stats]

    @stats.command()
    @commands.guild_only()
    async def add(self, ctx: Context, name: str):
        self._logger.debug(__name__, f'Received command stats add {ctx}: {name}')
        if not await self._client_utils.check_if_bot_is_ready_yet_and_respond(ctx):
            return

        if not self._permissions.is_member_technician(ctx.author):
            await self._message_service.send_ctx_msg(ctx, self._t.transform('common.no_permission_message'))
            self._logger.trace(__name__, f'Finished command stats add')
            return

        form = AddStatisticForm(stats, name, self._message_service, self._logger, self._t)
        self._logger.trace(__name__, f'Finished stats add command')
        self._logger.trace(__name__, f'Started stats command form')
        await ctx.interaction.response.send_modal(form)

    @stats.command()
    @commands.guild_only()
    async def edit(self, ctx: Context, name: str):
        self._logger.debug(__name__, f'Received command stats edit {ctx}: {name}')
        if not await self._client_utils.check_if_bot_is_ready_yet_and_respond(ctx):
            return

        if not self._permissions.is_member_technician(ctx.author):
            await self._message_service.send_ctx_msg(ctx, self._t.transform('common.no_permission_message'))
            self._logger.trace(__name__, f'Finished command stats edit')
            return

        try:
            statistic = stats.where(lambda s: s.name == name).single()
            form = AddStatisticForm(stats, name, self._message_service, self._logger, self._t, code=statistic.code, description=statistic.description)
            self._logger.trace(__name__, f'Finished stats edit command')
            self._logger.trace(__name__, f'Started stats command form')
            await ctx.interaction.response.send_modal(form)
        except Exception as e:
            self._logger.error(__name__, f'Cannot edit statistic {name}', e)
            await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.stats.edit.failed'))

    @edit.autocomplete('name')
    async def edit_autocomplete(self, interaction: discord.Interaction, current: str) -> TList[app_commands.Choice[str]]:
        return [app_commands.Choice(name=f'{statistic.name}: {statistic.description}', value=statistic.name) for statistic in stats]

    @stats.command()
    @commands.guild_only()
    async def remove(self, ctx: Context, name: str):
        self._logger.debug(__name__, f'Received command stats remove {ctx}: {name}')
        if not await self._client_utils.check_if_bot_is_ready_yet_and_respond(ctx):
            return

        if not self._permissions.is_member_technician(ctx.author):
            await self._message_service.send_ctx_msg(ctx, self._t.transform('common.no_permission_message'))
            self._logger.trace(__name__, f'Finished command stats remove')
            return

        try:
            stats.remove(stats.where(lambda s: s.name == name).single())
            self._logger.trace(__name__, f'Finished stats remove command')
        except Exception as e:
            self._logger.error(__name__, f'Cannot remove statistic {name}', e)
            await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.stats.edit.failed'))

    @remove.autocomplete('name')
    async def edit_autocomplete(self, interaction: discord.Interaction, current: str) -> TList[app_commands.Choice[str]]:
        return [app_commands.Choice(name=f'{statistic.name}: {statistic.description}', value=statistic.name) for statistic in stats]
